@@ -153,22 +153,6 @@ void ass_free_style(ASS_Track *track, int sid)
 
 // ==============================================================================================
 
-static void skip_spaces(char **str)
-{
-    char *p = *str;
-    while ((*p == ' ') || (*p == '\t'))
-        ++p;
-    *str = p;
-}
-
-static void rskip_spaces(char **str, char *limit)
-{
-    char *p = *str;
-    while ((p >= limit) && ((*p == ' ') || (*p == '\t')))
-        --p;
-    *str = p;
-}
-
 /**
  * \brief Set up default style
  * \param style style to edit to defaults
@@ -195,13 +179,6 @@ static void set_default_style(ASS_Style *style)
     style->MarginL = style->MarginR = style->MarginV = 20;
 }
 
-static uint32_t string2color(ASS_Library *library, char *p)
-{
-    uint32_t tmp;
-    (void) strtocolor(library, &p, &tmp, 0);
-    return tmp;
-}
-
 static long long string2timecode(ASS_Library *library, char *p)
 {
     unsigned h, m, s, ms;
@@ -267,7 +244,7 @@ static int numpad2align(int val)
 
 #define COLORVAL(name) \
 	} else if (strcasecmp(tname, #name) == 0) { \
-		target->name = string2color(track->library, token);
+		target->name = string2color(track->library, token, 0);
 
 #define INTVAL(name) ANYVAL(name,atoi)
 #define FPVAL(name) ANYVAL(name,ass_atof)
@@ -297,12 +274,7 @@ static char *next_token(char **str)
         *p = '\0';
         *str = p + 1;           // ',' found, str will point to the next char (beginning of the next token)
     }
-    --p;                        // end of current token
-    rskip_spaces(&p, start);
-    if (p < start)
-        p = start;              // empty token
-    else
-        ++p;                    // the first space character, or '\0'
+    rskip_spaces(&p, start);    // end of current token: the first space character, or '\0'
     *p = '\0';
     return start;
 }
@@ -607,9 +579,7 @@ static int process_info_line(ASS_Track *track, char *str)
     } else if (!strncmp(str, "Language:", 9)) {
         char *p = str + 9;
         while (*p && isspace(*p)) p++;
-        track->Language = malloc(3);
-        strncpy(track->Language, p, 2);
-        track->Language[2] = 0;
+        track->Language = strndup(p, 2);
     }
     return 0;
 }
@@ -697,6 +667,8 @@ static int decode_font(ASS_Track *track)
         goto error_decode_font;
     }
     buf = malloc(size / 4 * 3 + 2);
+    if (!buf)
+        goto error_decode_font;
     q = buf;
     for (i = 0, p = (unsigned char *) track->parser_priv->fontdata;
          i < size / 4; i++, p += 4) {
@@ -804,12 +776,6 @@ static int process_line(ASS_Track *track, char *str)
             break;
         }
     }
-
-    // there is no explicit end-of-font marker in ssa/ass
-    if ((track->parser_priv->state != PST_FONTS)
-        && (track->parser_priv->fontname))
-        decode_font(track);
-
     return 0;
 }
 
@@ -837,6 +803,9 @@ static int process_text(ASS_Track *track, char *str)
             break;
         p = q;
     }
+    // there is no explicit end-of-font marker in ssa/ass
+    if (track->parser_priv->fontname)
+        decode_font(track);
     return 0;
 }
 
@@ -849,6 +818,8 @@ static int process_text(ASS_Track *track, char *str)
 void ass_process_data(ASS_Track *track, char *data, int size)
 {
     char *str = malloc(size + 1);
+    if (!str)
+        return;
 
     memcpy(str, data, size);
     str[size] = '\0';
@@ -909,6 +880,8 @@ void ass_process_chunk(ASS_Track *track, char *data, int size,
     }
 
     str = malloc(size + 1);
+    if (!str)
+        return;
     memcpy(str, data, size);
     str[size] = '\0';
     ass_msg(track->library, MSGL_V, "Event at %" PRId64 ", +%" PRId64 ": %s",
@@ -995,6 +968,9 @@ static char *sub_recode(ASS_Library *library, char *data, size_t size,
 #endif
     }
 
+    if (icdsc == (iconv_t) (-1))
+        return NULL;
+
     {
         size_t osize = size;
         size_t ileft = size;
@@ -1005,6 +981,8 @@ static char *sub_recode(ASS_Library *library, char *data, size_t size,
         int clear = 0;
 
         outbuf = malloc(osize);
+        if (!outbuf)
+            goto out;
         ip = data;
         op = outbuf;
 
@@ -1018,7 +996,13 @@ static char *sub_recode(ASS_Library *library, char *data, size_t size,
             if (rc == (size_t) (-1)) {
                 if (errno == E2BIG) {
                     size_t offset = op - outbuf;
-                    outbuf = (char *) realloc(outbuf, osize + size);
+                    char *nbuf = realloc(outbuf, osize + size);
+                    if (!nbuf) {
+                        free(outbuf);
+                        outbuf = 0;
+                        goto out;
+                    }
+                    outbuf = nbuf;
                     op = outbuf + offset;
                     osize += size;
                     oleft += size;
@@ -1077,7 +1061,11 @@ static char *read_file(ASS_Library *library, char *fname, size_t *bufsize)
 
     ass_msg(library, MSGL_V, "File size: %ld", sz);
 
-    buf = malloc(sz + 1);
+    buf = sz < SIZE_MAX ? malloc(sz + 1) : NULL;
+    if (!buf) {
+        fclose(fp);
+        return NULL;
+    }
     assert(buf);
     bytes_read = 0;
     do {
@@ -1116,10 +1104,6 @@ static ASS_Track *parse_memory(ASS_Library *library, char *buf)
     for (i = 0; i < track->n_events; ++i)
         track->events[i].ReadOrder = i;
 
-    // there is no explicit end-of-font marker in ssa/ass
-    if (track->parser_priv->fontname)
-        decode_font(track);
-
     if (track->track_type == TRACK_TYPE_UNKNOWN) {
         ass_free_track(track);
         return 0;
@@ -1306,9 +1290,15 @@ long long ass_step_sub(ASS_Track *track, long long now, int movement)
 ASS_Track *ass_new_track(ASS_Library *library)
 {
     ASS_Track *track = calloc(1, sizeof(ASS_Track));
+    if (!track)
+        return NULL;
     track->library = library;
     track->ScaledBorderAndShadow = 1;
     track->parser_priv = calloc(1, sizeof(ASS_ParserPriv));
+    if (!track->parser_priv) {
+        free(track);
+        return NULL;
+    }
     return track;
 }
 
